#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>

#include "asim/provides/connected_application.h"
#include "asim/provides/stats_device.h"
//#include "asim/provides/SndfileWavUtil.h"

#include "asim/rrr/client_stub_AUDIOCORERRR.h"

using namespace std;

#define NUM_VOICES 2

pthread_mutex_t CONNECTED_APPLICATION_CLASS::lock;
pthread_cond_t  CONNECTED_APPLICATION_CLASS::cond;
sem_t CONNECTED_APPLICATION_CLASS::throttle;


// constructor
CONNECTED_APPLICATION_CLASS::CONNECTED_APPLICATION_CLASS(VIRTUAL_PLATFORM vp) :
   
  clientStub(new AUDIOCORERRR_CLIENT_STUB_CLASS(this))
{
  printf("SendSide Constructed\n");
}

// destructor
CONNECTED_APPLICATION_CLASS::~CONNECTED_APPLICATION_CLASS()
{
}

// init
void
CONNECTED_APPLICATION_CLASS::Init()
{

  printf("InitSend Side\n");
  pthread_mutex_init(&lock, NULL);
  pthread_cond_init(&cond, NULL);
  sem_init(&throttle, 0, 256);

  // enable stats 
  //STATS_DEVICE_SERVER_CLASS::GetInstance()->SetupStats();
}

//This isn't called.
void
CONNECTED_APPLICATION_CLASS::UpdateSemaphore()
{
  printf("UpdateSemaphore\n");
  sem_post(&throttle);
}

void
CONNECTED_APPLICATION_CLASS::EndSimulation()
{
  printf("EndSimulation Called\n");
  fflush(stdout);
  pthread_mutex_lock(&lock);
  // Do something about the race occuring here
  pthread_cond_signal(&cond); 
  pthread_mutex_unlock(&lock);
  printf("EndSimulation done\n");
  fflush(stdout);  
}

bool CONNECTED_APPLICATION_CLASS::areAllFilesComplete(int numVoices, bool isFini[])
{
  bool result = false;
  for (int n = 0; n < numVoices; n++)
      result |= isFini[n];
  return result;
}

// main
void
CONNECTED_APPLICATION_CLASS::Main()
{

  //rlm: two files
  FILE *inputFiles[NUM_VOICES];
  UINT16 sample;

  bool fileFini[NUM_VOICES];
  UINT16 channel[NUM_VOICES];
  for (int n = 0; n < NUM_VOICES; n++)
    {
      fileFini[n] = false;
      channel[n] = n;
    }

  fflush(stdout); 

  // Convert input wav to pcm
 
  printf("word!\n");

  for (int i = 0; i < NUM_VOICES; i++)
    {
      char filename[256];
      sprintf(filename, "input%d", i);
      char wavname[256];
      strcpy(wavname, filename);
      strcat(wavname, ".wav");
      char pcmname[256];
      strcpy(pcmname, filename);
      strcat(pcmname, ".pcm");

      printf("Opening audio file %s and creating %s.\n", wavname, pcmname);	   
      inputFiles[i] = fopen(wavname, "r");

      generate_pcm(wavname, pcmname);  

      assert(inputFiles[i]);
    }

  //init processor 
  int sleepCount = 0;
  int result = 0;
  int count = 0;

  int numCoreFini = 0; 

  printf("SendSide Main\n");
  fflush(stdout); 

  // Convert input wav to pcm
 
  sleep(2);
  printf("main:samples about to start sending %d\n", count);  

  //rlm: here we read both files. later refactor into a function.
  // also, this will terminate when the FIRST file reaches its end.
  while(!areAllFilesComplete(NUM_VOICES, fileFini))
    { 
     sem_wait(&throttle);
 
     if(count%1000 == 0)
     printf("main: %d\n", count);
     count++;
    
    //rlm: two files.
     for (int i = 0; i < NUM_VOICES; i++)
       {
	 if (!fileFini[i])
	   {
	     if (fread(&sample, 2, 1, inputFiles[i]))
	       clientStub->SendUnprocessedStream((UINT32)channel[i] , Data,(UINT32)sample);
	     else{
	       fileFini[i] = true;
	       numCoreFini++;
	       if (numCoreFini >= NUM_VOICES)  //  Need lock on last eof
		 // Need to put lock here to prevent potential race condition
		 pthread_mutex_lock(&lock);
	       clientStub->SendUnprocessedStream((UINT32)channel[i], EndOfFile,0);
	     }
	   }
       }
    }

  printf("main: out of loop\n");
  printf("main: wait for end of file\n");

  pthread_cond_wait(&cond, &lock);
  pthread_mutex_unlock(&lock);

  printf("main: last data out\n");

  // Convert input wav to pcm
  // this part is ok because there is always only one input file.
  generate_wav("out_hw.pcm","input0.wav","out_hw.wav");

  printf("generate wav done\n");

  //  printf("If you missed it, core %s", coreFin ? "PASSED\n" : "FAILED\n");
  // Dump the stats file

  STATS_DEVICE_SERVER_CLASS::GetInstance()->DumpStats();
  STATS_DEVICE_SERVER_CLASS::GetInstance()->EmitFile();

  fflush(stdout);
  //  exit(0);
}
